@@ -1,8 +1,8 @@
 name = 'ldf_adapter'
-# vim: foldmethod=indent : tw=100
+# vim: tw=100 foldmethod=indent
 # pylint: disable=invalid-name, superfluous-parens
-# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
-# pylint: disable=missing-docstring, too-few-public-methods
+# pylint: disable=redefined-outer-name, logging-not-lazy, logging-format-interpolation, logging-fstring-interpolation
+# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
 
 import logging
 from collections import Mapping
@@ -14,6 +14,8 @@ import urllib
 import regex
 from unidecode import unidecode
 
+from . import logsetup
+# from ldf_adapter.logsetup import jsonlogger
 from . import eduperson
 
 from . import backend
@@ -53,6 +55,7 @@ class User:
         self.service_groups = [backend.Group(grp) for grp in self.data.groups]
 
         if CONFIG.get('ldf_adapter', 'backend_supports_preferring_existing_user', fallback = False):
+            logger.debug("trying to update user from existing")
             if self.service_user.exists():
                 self.update_username_from_existing()
 
@@ -163,11 +166,11 @@ class User:
         if self.service_user.exists():
             username = self.service_user.get_username()
         try:
-            logger.info(F"Incoming request to reach '{target}' for user '{self.data.email}' ({self.data.unique_id}) username: {username}")
+            logger.info(F"Incoming request to reach '{target}' for user with email: '{self.data.email}' ({self.data.unique_id}) username: {username}")
         except AttributeError:
-            logger.info(F"Incoming request to reach '{target}' for user '{self.data.full_name}' ({self.data.unique_id}) username: {username}")
+            logger.info(F"Incoming request to reach '{target}' for user with name: '{self.data.full_name}' ({self.data.unique_id}) username: {username}")
         except AttributeError:
-            logger.info(F"Incoming request to reach '{target}' for user '{self.data.unique_id}' username: {username}")
+            logger.info(F"Incoming request to reach '{target}' for user with unique_id: '{self.data.unique_id}' username: {username}")
 
         if target == 'deployed':
             if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
@@ -193,8 +196,8 @@ class User:
             return self.resume()
         elif target == 'suspended':
             return self.suspend()
-        elif target == 'expired':
-            return self.expire()
+        elif target == 'limited':
+            return self.limit()
         else:
             raise ValueError(f"Invalid target state: {target}")
 
@@ -235,14 +238,15 @@ class User:
 
         Return a NotDeployed result with a message saying if the user previously existed.
         """
+        username = self.service_user.get_username()
         was_removed = self.ensure_dosent_exist()
 
         what_changed = ''
         if was_removed:
-            what_changed += F"User '{self.data.username}' was removed."
+            what_changed += F"User '{username} ({self.data.unique_id})' was removed."
         else:
             what_changed += F"No user for '{self.data.unique_id}' existed. "+\
-                            F"User '{self.data.username}' was not changed"
+                            F"User '{username}' was not changed"
 
         return NotDeployed(message=what_changed)
 
@@ -264,36 +268,52 @@ class User:
         return Status(state, message=what_changed)
 
     def resume(self):
-        """Ensure that a suspended or expired user is active again in state 'deployed'.
+        """Ensure that a suspended user is active again in state 'deployed'.
 
         Return a Status result with a message describing what was done.
         """
         was_resumed = self.ensure_resumed()
         what_changed = ''
+        state = self.get_status().state
         if was_resumed:
             # FIXME: I'm not sure if this ought to be self.data.username
             what_changed += F"User '{self.data.unique_id}' was resumed."
-            state = "deployed"
         else:
-            state = self.get_status().state
             what_changed += F"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                             F"User was not changed."
         return Status(state, message=what_changed)
 
-    def expire(self):
-        """Ensure that a user is expired.
+    def limit(self):
+        """Ensure that a user has limited access.
 
         Return a Status result with a message describing what was done.
         """
-        was_expired = self.ensure_expired()
+        was_limited = self.ensure_limited()
         what_changed = ''
-        if was_expired:
+        if was_limited:
             # FIXME: I'm not sure if this ought to be self.data.username
-            what_changed += F"User '{self.data.unique_id}' was expired."
-            state = "expired"
+            what_changed += F"User '{self.data.unique_id}' was limited."
+            state = "limited"
         else:
             state = self.get_status().state
-            what_changed += F"Expiring user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
+            what_changed += F"Limiting user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
+                            F"User was not changed."
+        return Status(state, message=what_changed)
+
+    def unlimit(self):
+        """Ensure that a limited user is active again in state 'deployed'.
+
+        Return a Status result with a message describing what was done.
+        """
+        was_unlimited = self.ensure_unlimited()
+        what_changed = ''
+        if was_unlimited:
+            # FIXME: I'm not sure if this ought to be self.data.username
+            what_changed += F"User '{self.data.unique_id}' was unlimited."
+            state = "deployed"
+        else:
+            state = self.get_status().state
+            what_changed += F"Resuming user '{self.data.unique_id}' was not possible from the '{state}' state. "+\
                             F"User was not changed."
         return Status(state, message=what_changed)
 
@@ -316,7 +336,7 @@ class User:
         +--------------+-----------------------------------------------------------------+-----------------+
         | pending      | The creation of the user is pending                             | Optional        |
         +--------------+-----------------------------------------------------------------+-----------------+
-        | expired      | The user was expired, typically after being idle for some time  | Optional        |
+        | limited      | The user was limited, typically after being idle for some time  | Optional        |
         +--------------+-----------------------------------------------------------------+-----------------+
         | unknown      | We don't know the status, but at least the user is not deployed | Mandatory       |
         +--------------+-----------------------------------------------------------------+-----------------+
@@ -326,7 +346,7 @@ class User:
         try:
             if not self.service_user.exists():
                 return Status("not_deployed", message=msg)
-            msg=F"username {self.data.username}"
+            msg=F"username {self.service_user.get_username()}"
             if hasattr(self.service_user, "is_rejected"):
                 if self.service_user.is_rejected():
                     return Status("rejected", message=msg)
@@ -336,9 +356,9 @@ class User:
             if hasattr(self.service_user, "is_pending"):
                 if self.service_user.is_pending():
                     return Status("pending", message=msg)
-            if hasattr(self.service_user, "is_expired"):
-                if self.service_user.is_expired():
-                    return Status("expired", message=msg)
+            if hasattr(self.service_user, "is_limited"):
+                if self.service_user.is_limited():
+                    return Status("limited", message=msg)
             return Status("deployed", message=msg)
         except Exception as e:
             logger.error(F'User {self.data.unique_id} is in an undefined state.: {e}')
@@ -362,10 +382,11 @@ class User:
         if is_new_user:
             unique_id= self.data.unique_id
             logger.info(F'Creating user for "{unique_id}"')
+            username = self.data.username
 
             # Raise question in case of existing username in case we're interactive
             if CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False): # interactive
-                if self.service_user.name_taken():
+                if self.service_user.name_taken(username):
                     logger.info(F'Username "{username}" is already taken, asking user to pick a new one')
                     raise Question(
                         name='username',
@@ -373,18 +394,19 @@ class User:
                     )
 
             else: # non-interactive
-                fng = FriendlyNameGenerator(self.data)
                 try:
-                    logger.info(F"initial try: {self.data.username}")
+                    logger.info(F"initial try: {username}")
                 except AttributeError:
                     pass
-                while self.service_user.name_taken():
-                    # FIXME: This may as well be data.username!! or a new  set_username
-                    self.service_user.name = fng.suggest_name(self.data.username)
-                logger.info(F'                             Using: {self.data.username}')
-                if self.service_user.name is None:
+                fng = FriendlyNameGenerator(self.data)
+                proposed_name = username if username is not None else fng.suggest_name()
+                while self.service_user.name_taken(proposed_name):
+                    proposed_name = fng.suggest_name()
+                logger.info(F'Using: {proposed_name}')
+                if proposed_name is None:
                     raise Rejection(message=F"I cannot create usernames. "
                                     F"The list of tried ones is: {', '.join(fng.tried_names())}.")
+                self.service_user.set_username(proposed_name)
             self.service_user.create()
         else: # The user exists
             # Update service_user.name if unique_id already points to a username:
@@ -405,7 +427,7 @@ class User:
                 if hasattr(self.service_user, 'set_username'):
                     logger.debug(F"Setting username to {existing_username} ({self.data.unique_id})")
                     self.service_user.set_username(existing_username)
-                logger.info(F'Found existing username: {existing_username}')
+                logger.info(F'Found an existing username: {existing_username}')
         except AttributeError:
             # the currently used service_user class has to method get_username
             existing_username = None
@@ -420,9 +442,15 @@ class User:
         if self.service_user.exists():
             self.service_user.username = self.service_user.get_username()
             logger.info(F"Deleting user '{self.service_user.username}' ({self.data.unique_id})")
-            self.service_user.username = self.service_user.get_username()
-            self.service_user.delete()
-            self.service_user.uninstall_ssh_keys()
+            # bwIDM requires prior removal of the user, because ssh-key removal triggers an 
+            # asyncronous process. If user is removed during that, the user might be only partially
+            # removed...
+            if CONFIG.get('ldf_adapter','backend', fallback="") == 'bwidm':
+                self.service_user.delete()
+                self.service_user.uninstall_ssh_keys()
+            else:
+                self.service_user.uninstall_ssh_keys()
+                self.service_user.delete()
             return True
         else:
             logger.debug(F'No user for {self.data.unique_id} did exist. Nothing to do.')
@@ -433,23 +461,23 @@ class User:
         Return True if the user has been suspended.
         """
         status = self.get_status()
-        if status.state in ["deployed", "expired"]:
+        if status.state in ["deployed", "limited"]:
             if hasattr(self.service_user, 'suspend'):
                 self.service_user.suspend()
                 return True
         logger.debug(F'User {self.data.unique_id} in state {status.state}. Suspending not allowed.')
         return False
 
-    def ensure_expired(self):
-        """Ensure that a user is expired.
-        Return True if setting the user is expired.
+    def ensure_limited(self):
+        """Ensure that a user has limited access.
+        Return True if setting the user is limited.
         """
         status = self.get_status()
         if status.state == "deployed":
-            if hasattr(self.service_user, 'expire'):
-                self.service_user.expire()
+            if hasattr(self.service_user, 'limit'):
+                self.service_user.limit()
                 return True
-        logger.debug(F'User {self.data.unique_id} in state {status.state}. Expiring not allowed.')
+        logger.debug(F'User {self.data.unique_id} in state {status.state}. Limiting not allowed.')
         return False
 
     def ensure_resumed(self):
@@ -457,13 +485,25 @@ class User:
         Return True is the user
         """
         status = self.get_status()
-        if status.state in ["suspended", "expired"]:
+        if status.state == "suspended":
             if hasattr(self.service_user, 'resume'):
                 self.service_user.resume()
                 return True
         logger.debug(F'User {self.data.unique_id} in state {status.state}. Resuming not allowed.')
         return False
 
+    def ensure_unlimited(self):
+        """Ensure that a user is not limited anymore.
+        Return True is the user
+        """
+        status = self.get_status()
+        if status.state == "limited":
+            if hasattr(self.service_user, 'unlimit'):
+                self.service_user.unlimit()
+                return True
+        logger.debug(F'User {self.data.unique_id} in state {status.state}. Unlimit not allowed.')
+        return False
+
     def ensure_groups_exist(self):
         """Ensure that all the necessary groups exist.
 
@@ -540,6 +580,14 @@ class UserInfo(Mapping):
         self.credentials = data['user'].get('credentials', {})
         self.allow_question = CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False)
 
+    @property
+    @lru_cache(maxsize=None)
+    def size(self):
+        the_size = 0
+        for x in self.userinfo.keys():
+            the_size+=1
+        return the_size
+
     @property
     @lru_cache(maxsize=None)
     def unique_id(self):
@@ -594,7 +642,6 @@ class UserInfo(Mapping):
         if sub != self.userinfo['sub']:
             logger.warning("sub '{}' changed to '{}' for BWIDM compatibilty".format(
                 self.userinfo['sub'], sub))
-
         return sub
 
     def _iss_masked_for_bwidm_eppn(self):
@@ -614,23 +661,28 @@ class UserInfo(Mapping):
             if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
                 logger.warning("Issuer '{}' changed to '{}' for BWIDM compatibilty".format(
                     stripped_iss, iss))
-
         return iss
 
     @property
     @lru_cache(maxsize=None)
     def username(self):
-        """Return the user's preferred username, or ask for one if none was provided."""
-        # FIXME: This function is called, even when there is already a local user with an existing
-        # name.  SImply not having a "preferred_username" does not mean that we have to bother the user!!!
+        """Return the user's name, this may be:
+            - preferred_username if that was provided
+            - if in interactive mode, we prompt the user
+            - otherwise, None"""
+        # FIXME: If this function is called, even when there is already a local user with an existing
+        # name, this is a bug in the flow, that MUST be fixed
+        #
+        # Simply not having a "preferred_username" does not mean that we have to bother the user!!!
+        # TODO (DG): does this still need to be fixed? it should work => test 
+
         if self.allow_question:
             return self.value_or_ask(
                 self.userinfo.get('preferred_username'), 'username',
                 'You have not set a global username preference. Please enter your preferred username.',
                 self.allow_question
             )
-        fng = FriendlyNameGenerator(self)
-        return fng.suggest_name()
+        return self.userinfo.get('preferred_username', None)
 
     @property
     @lru_cache(maxsize=None)
@@ -805,7 +857,7 @@ class UserInfo(Mapping):
     @lru_cache(maxsize=None)
     def preferred_username(self):
         """Return the prefrred username of the user."""
-        return self.userinfo.get('preferred_username', [])
+        return self.userinfo.get('preferred_username', None)
 
 
 
