@@ -11,6 +11,8 @@ import subprocess
 from subprocess import CalledProcessError
 from pathlib import Path
 from os import chown
+from datetime import date, datetime
+import json
 import logging
 
 import regex
@@ -31,8 +33,14 @@ class User:
                 `unique_id`  stored in gecos, used to find the user
                 `ssh_keys`
         """
-        self.name = make_shadow_compatible(userinfo.username)
         self.unique_id = userinfo.unique_id
+        logger.debug(F"backend processing: {userinfo.unique_id}")
+        if self.exists():
+            logger.debug(F"This user does actually exist. The name is: {self.get_username()}")
+            self.set_username(self.get_username())
+        else:
+            self.set_username(userinfo.username)
+
         self.ssh_keys = [key['value'] for key in userinfo.ssh_keys]
         self.primary_group = Group(userinfo.primary_group)
         self.credentials = {}
@@ -51,6 +59,29 @@ class User:
     def is_suspended(self):
         """Optional, only if the backend supports it.
         Inform the user whether a user was suspended (e.g. due to a security incident)"""
+        if self.exists():
+            options = ['-i', '-l']
+            try:
+                result = subprocess.run(['chage'] + options + [self.name],
+                                        capture_output=True, check=True)
+            except CalledProcessError as e:
+                msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
+                logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
+                raise Failure(message=F"Cannot get info for user: {msg or '<no output>'}")
+            try:
+                pattern = regex.compile(r'Account expires\s+: (.*)')
+                match = pattern.search(result.stdout.decode("utf-8"))
+                if match:
+                    expiration_date = match.group(1)
+                    if expiration_date == 'never':
+                        return False
+                    expiration_date_sec = int(
+                        datetime.fromisoformat(expiration_date).strftime('%s'))
+                    if expiration_date_sec - datetime.now().timestamp() <= 0:
+                        return True
+            except Exception as e:
+                logger.error(e)
+                raise Failure(message=F"Could not get: {e or '<no output>'}")
         return False
 
     def is_pending(self):
@@ -58,16 +89,26 @@ class User:
         Inform the user whether his creation is pending"""
         return False
 
-    def is_expired(self):
+    def is_limited(self):
         """Optional, only if the backend supports it.
-        Inform the user whether his creation is pending"""
-        return False
+        Inform the user whether a user has limited access"""
+        # if shell is nologin
+        try:
+            shell = self.__passwd_entry.get('shell', None)
+            return (
+                shell == '/sbin/nologin' or
+                shell == '/usr/bin/nologin' or
+                shell == '/bin/nologin'
+                )
+        except KeyError:
+            return False
 
-    def name_taken(self):
-        """Check if a username is already taken"""
-        x = self.name in [entry['login'] for entry in User.__all_passwd_entries('login').values()]
-        logger.debug(F"name_taken: {x}")
-        return self.name in [entry['login'] for entry in User.__all_passwd_entries('login').values()]
+    def name_taken(self, name):
+        """Check if a username is already taken by *another* user"""
+        name = make_shadow_compatible(name)
+        taken = name in [entry['login'] for entry in User.__all_passwd_entries('login').values()]
+        logger.debug(F"name_taken: {taken}")
+        return taken and name != self.get_username()
 
     def get_username(self):
         """Return username based on unique_id"""
@@ -78,10 +119,13 @@ class User:
         except KeyError:
             return None
 
+    def set_username(self, username):
+        """Set local username on the service."""
+        self.name = make_shadow_compatible(username)
+
     def create(self):
         logger.debug(F"creating user: {self.name} - {self.unique_id} ")
         try:
-            # TODO this should consider self.primary_group
             shell = CONFIG['backend.local_unix'].get('shell', '/bin/sh')
             subprocess.run(['useradd', '--comment', self.unique_id,
                             '-g', self.primary_group.name,
@@ -130,6 +174,37 @@ class User:
             logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
             raise Failure(message=F"Cannot modify user: {msg or '<no output>'}")
 
+    def __expire(self, expiration_date=datetime.today().strftime("%Y-%m-%d")):
+        options = ['-E', expiration_date]
+        try:
+            subprocess.run(['chage'] + options + [self.name],
+                           capture_output=True, check=True)
+        except CalledProcessError as e:
+            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
+            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
+            raise Failure(message=F"Cannot set expiration date for user: {msg or '<no output>'}")
+
+    def suspend(self):
+        self.__expire(datetime.today().strftime("%Y-%m-%d"))
+
+    def resume(self):
+        self.__expire('-1')
+
+    def __set_shell(self, shell):
+        try:
+            subprocess.run(['usermod', '-s', shell, self.name], check=True)
+        except CalledProcessError as e:
+            msg = (e.stderr or e.stdout or b'').decode('utf-8').strip()
+            logger.error('Error executing \'{}\': {}'.format(' '.join(e.cmd), msg or "<no output>"))
+            raise Failure(message=F"Cannot modify user: {msg or '<no output>'}")
+
+    def limit(self):
+        self.__set_shell('/sbin/nologin')
+
+    def unlimit(self):
+        shell = CONFIG['backend.local_unix'].get('shell', '/bin/sh')
+        self.__set_shell(shell)
+
     def install_ssh_keys(self):
         try:
             self.__authorized_keys.parent.mkdir(parents=True, exist_ok=True)
@@ -288,8 +363,30 @@ def make_shadow_compatible(orig_word):
     # aber nur *ganz* am Ende ...
     word = regex.sub(r'[^-0-9_a-z]', '_', word[:-1]) + regex.sub(r'[^-0-9_a-z$]', '_', word[-1])
 
+
+    # usernames and group names can only be 32 characters long.
+    # My fix is to remove characters a) after the first '_' if there is one.
+    #                                b) from the beginning if there is none
+    # Also adds two dots as an indicator for where the shortening took place
+
+    excess_chars = len(word) - 32
+    if excess_chars > 0:
+
+        if len(word.split('_')) == 1: # no '_' found:
+            word = '..' + word[excess_chars+2:] 
+            logger.warning(F"shortened {orig_word} to {word}")
+
+        if len(word.split('_')) > 1: # at least one '_' found:
+            fragments = word.split('_')
+            if len(fragments[1]) > excess_chars: # we're fine, we can cut excess chars from fragments alone
+                fragments[1]=".."+fragments[1][excess_chars+2:]
+                word = '_'.join(fragments)
+            else:
+                logger.error(F"User or group name is too long")
+                raise(ValueError)
+            logger.warning(F"shortened {orig_word} to {word}")
+
     if word != orig_word:
         logger.warning("Name '{}' changed to '{}' for shadow compatibilty".format(orig_word, word))
 
     return word
-
